@@ -2,8 +2,8 @@
 
 #
 #
-# This code is from the project https://github.com/inaz2/proxy2, with some
-# minor modifications.
+# This code originated from the project https://github.com/inaz2/proxy2 but has since
+# been modified extensively.
 #
 #
 
@@ -11,31 +11,16 @@ import base64
 import re
 import socket
 import ssl
-import sys
 import threading
 import urllib.parse
 from http.client import HTTPConnection, HTTPSConnection
-from http.server import BaseHTTPRequestHandler, HTTPServer
-from socketserver import ThreadingMixIn
+from http.server import BaseHTTPRequestHandler
 
 from . import cert, socks
 
 
-class ThreadingHTTPServer(ThreadingMixIn, HTTPServer):
-    address_family = socket.AF_INET6
-    daemon_threads = True
-
-    def handle_error(self, request, client_address):
-        # surpress socket/ssl related errors
-        cls, e = sys.exc_info()[:2]
-        if issubclass(cls, socket.error) or issubclass(cls, ssl.SSLError):
-            pass
-        else:
-            return HTTPServer.handle_error(self, request, client_address)
-
-
 class ProxyRequestHandler(BaseHTTPRequestHandler):
-    admin_path = 'http://proxy2'
+    admin_path = None
     # Path to the directory used to store the generated certificates.
     # Subclasses can override certdir
     certdir = cert.CERTDIR
@@ -53,7 +38,10 @@ class ProxyRequestHandler(BaseHTTPRequestHandler):
 
         certpath = cert.generate(self.path.split(':')[0], self.certdir)
 
-        with ssl.wrap_socket(self.connection, keyfile=cert.CERTKEY, certfile=certpath, server_side=True) as conn:
+        with ssl.wrap_socket(self.connection,
+                             keyfile=cert.CERTKEY,
+                             certfile=certpath,
+                             server_side=True) as conn:
             self.connection = conn
             self.rfile = conn.makefile('rb', self.rbufsize)
             self.wfile = conn.makefile('wb', self.wbufsize)
@@ -64,10 +52,11 @@ class ProxyRequestHandler(BaseHTTPRequestHandler):
         else:
             self.close_connection = True
 
-    def do_GET(self):
-        if self.path.startswith(self.admin_path):
+    def proxy_request(self):
+        if self.admin_path and self.path.startswith(self.admin_path):
             self.admin_handler()
             return
+        
         req = self
         content_length = int(req.headers.get('Content-Length', 0))
         req_body = self.rfile.read(content_length) if content_length else None
@@ -93,12 +82,11 @@ class ProxyRequestHandler(BaseHTTPRequestHandler):
         assert scheme in ('http', 'https')
         if netloc:
             req.headers['Host'] = netloc
-        setattr(req, 'headers', self.filter_headers(req.headers))
+        setattr(req, 'headers', self._filter_headers(req.headers))
 
         origin = (scheme, netloc)
-        conn = None
         try:
-            conn = self.create_connection(origin)
+            conn = self._create_connection(origin)
             conn.request(self.command, path, req_body, dict(req.headers))
             res = conn.getresponse()
 
@@ -115,9 +103,6 @@ class ProxyRequestHandler(BaseHTTPRequestHandler):
                 del self.tls.conns[origin]
             self.send_error(502)
             return
-        finally:
-            if conn and not self.websocket:
-                conn.close()
 
         res_body_modified = self.response_handler(req, req_body, res, res_body)
         if res_body_modified is False:
@@ -128,7 +113,7 @@ class ProxyRequestHandler(BaseHTTPRequestHandler):
             del res.headers['Content-length']
             res.headers['Content-Length'] = str(len(res_body))
 
-        setattr(res, 'headers', self.filter_headers(res.headers))
+        setattr(res, 'headers', self._filter_headers(res.headers))
 
         self.send_response(res.status, res.reason)
 
@@ -142,11 +127,12 @@ class ProxyRequestHandler(BaseHTTPRequestHandler):
         self.wfile.flush()
 
         if self.websocket:
-            self.handle_websocket(conn.sock)
-        else:
+            self._handle_websocket(conn.sock)
+            self.close_connection = True
+        elif not self._keepalive():
             self.close_connection = True
 
-    def create_connection(self, origin):
+    def _create_connection(self, origin):
         scheme, netloc = origin
 
         if origin not in self.tls.conns:
@@ -158,7 +144,7 @@ class ProxyRequestHandler(BaseHTTPRequestHandler):
 
             if scheme == 'https':
                 connection = ProxyAwareHTTPSConnection
-                if not self.server.options.get('verify_ssl', True):
+                if not self.server.options.get('verify_ssl', False):
                     kwargs['context'] = ssl._create_unverified_context()
             else:
                 connection = ProxyAwareHTTPConnection
@@ -167,14 +153,15 @@ class ProxyRequestHandler(BaseHTTPRequestHandler):
 
         return self.tls.conns[origin]
 
-    do_HEAD = do_GET
-    do_POST = do_GET
-    do_PUT = do_GET
-    do_DELETE = do_GET
-    do_OPTIONS = do_GET
-    do_PATCH = do_GET
+    do_HEAD = proxy_request
+    do_POST = proxy_request
+    do_GET = proxy_request
+    do_PUT = proxy_request
+    do_DELETE = proxy_request
+    do_OPTIONS = proxy_request
+    do_PATCH = proxy_request
 
-    def filter_headers(self, headers):
+    def _filter_headers(self, headers):
         # http://tools.ietf.org/html/rfc2616#section-13.5.1
         hop_by_hop = (
             'keep-alive',
@@ -213,11 +200,11 @@ class ProxyRequestHandler(BaseHTTPRequestHandler):
 
         return headers
 
-    def handle_one_request(self):
-        if not self.websocket:
-            super().handle_one_request()
+    def _keepalive(self):
+        return self.server.options.get('connection_keep_alive', True) \
+               and self.headers.get('Connection', '').lower() != 'close'
 
-    def handle_websocket(self, server_sock):
+    def _handle_websocket(self, server_sock):
         self.connection.settimeout(None)
         server_sock.settimeout(None)
 
@@ -255,26 +242,50 @@ class ProxyRequestHandler(BaseHTTPRequestHandler):
 
         t.join()
 
-    def send_cacert(self):
-        with open(cert.CACERT, 'rb') as f:
-            data = f.read()
+    def handle_one_request(self):
+        if not self.websocket:
+            super().handle_one_request()
 
-        self.send_response(200, 'OK')
-        self.send_header('Content-Type', 'application/x-x509-ca-cert')
-        self.send_header('Content-Length', len(data))
-        self.send_header('Connection', 'close')
-        self.end_headers()
-        self.wfile.write(data)
+    def finish(self):
+        for conn in self.tls.conns.values():
+            if conn:
+                conn.close()
+        super().finish()
 
     def request_handler(self, req, req_body):
+        """Hook method that subclasses should override to process a request.
+
+        Args:
+            req: A ProxyRequestHandler instance.
+            req_body: The request body as bytes.
+        """
         pass
 
     def response_handler(self, req, req_body, res, res_body):
+        """Hook method that subclasses should override to process a response.
+
+        Args:
+            req: The original request - a ProxyRequestHandler instance.
+            req_body: The request body as bytes.
+            res: The response (a http.client.HTTPResponse instance) that corresponds to the
+                request.
+            res_body: The response body as bytes.
+        """
         pass
 
     def admin_handler(self):
-        if self.path == 'http://proxy2.test/':
-            self.send_cacert()
+        """Subclasses should override this to process administration requests.
+
+        Administration requests are requests targeted at the proxy server itself
+        rather than a remote server.
+
+        Note that subclasses must set the admin_path class-level attribute to
+        a URL prefix that identifies administration requests. For example:
+            admin_path = 'http://myserver'
+        This method will then fire for any request with a URL path that starts
+        with http://myserver...
+        """
+        pass
 
     def log_error(self, format_, *args):
         # suppress "Request timed out: timeout('timed out',)"
